@@ -1,4 +1,6 @@
 #!/usr/bin/bash
-rm dist/*
+./build_docs.sh
+
+rm -f dist/*
 python setup.py sdist bdist_wheel
 twine check dist/*
