@@ -1,16 +1,17 @@
 """
 .. include:: ./documentation.md
 """
-from .serie import *
+from .gpr.config import *
+from .util import *
+
+from .transformer import *
 from .data import *
 from .dataset import *
+from .init import *
 from .model import *
 from .models.sm import *
 from .models.mosm import *
 from .models.csm import *
 from .models.sm_lmc import *
 from .models.conv import *
-
-from .plot import *
-from .errors import *
-from .gpr.config import *
+from .models.mohsm import *
