@@ -4,9 +4,9 @@ import numpy as np
 import pandas as pd
 import matplotlib.pyplot as plt
 
-from .data import Data
+from .data import Data, _is_iterable
 
-def LoadCSV(filename, x_col=0, y_col=1, name=None, **kwargs):
+def LoadCSV(filename, x_col=0, y_col=1, y_err_col=None, name=None, **kwargs):
     """
     LoadCSV loads a dataset from a given CSV file. It loads in `x_col` as the names of the input dimension columns, and `y_col` as the names of the output columns.
 
@@ -29,9 +29,9 @@ def LoadCSV(filename, x_col=0, y_col=1, name=None, **kwargs):
 
     df = pd.read_csv(filename, **kwargs)
 
-    return LoadDataFrame(df, x_col, y_col, name)
+    return LoadDataFrame(df, x_col, y_col, y_err_col, name)
 
-def LoadDataFrame(df, x_col=0, y_col=1, name=None):
+def LoadDataFrame(df, x_col=0, y_col=1, y_err_col=None, name=None):
     """
     LoadDataFrame loads a DataFrame from pandas. It loads in `x_col` as the names of the input dimension columns, and `y_col` the names of the output columns.
 
@@ -39,6 +39,7 @@ def LoadDataFrame(df, x_col=0, y_col=1, name=None):
         df (pandas.DataFrame): The pandas DataFrame.
         x_col (int, str, list of int or str): Names or indices of X column(s) in DataFrame.
         y_col (int, str, list of int or str): Names or indices of Y column(s) in DataFrame.
+        y_err_col (int, str, list of int or str): Names or indices of Y error column(s) in DataFrame.
         name (str, list of str): Name or names of data channels.
 
     Returns:
@@ -50,20 +51,35 @@ def LoadDataFrame(df, x_col=0, y_col=1, name=None):
         <mogptk.dataset.DataSet at ...>
     """
 
+    if _is_iterable(x_col):
+        x_col = list(x_col)
+    if _is_iterable(y_col):
+        y_col = list(y_col)
     if (not isinstance(x_col, list) or not all(isinstance(item, int) for item in x_col) and not all(isinstance(item, str) for item in x_col)) and not isinstance(x_col, int) and not isinstance(x_col, str):
         raise ValueError("x_col must be integer, string or list of integers or strings")
     if (not isinstance(y_col, list) or not all(isinstance(item, int) for item in y_col) and not all(isinstance(item, str) for item in y_col)) and not isinstance(y_col, int) and not isinstance(y_col, str):
         raise ValueError("y_col must be integer, string or list of integers or strings")
-
     if not isinstance(x_col, list):
         x_col = [x_col]
     if not isinstance(y_col, list):
         y_col = [y_col]
 
+    if y_err_col is not None:
+        if _is_iterable(y_err_col):
+            y_col = list(y_err_col)
+        if (not isinstance(y_err_col, list) or not all(isinstance(item, int) for item in y_err_col) and not all(isinstance(item, str) for item in y_err_col)) and not isinstance(y_err_col, int) and not isinstance(y_err_col, str):
+            raise ValueError("y_err_col must be integer, string or list of integers or strings")
+        if not isinstance(y_err_col, list):
+            y_err_col = [y_err_col]
+        if len(y_col) != len(y_err_col):
+            raise ValueError("y_err_col and y_col must be of the same length")
+
     if name is None:
         name = [None] * len(y_col)
     else:
-        if not isinstance(name, list):
+        if _is_iterable(name):
+            name = list(name)
+        else:
             name = [name]
         if len(y_col) != len(name):
             raise ValueError("y_col and name must be of the same length")
@@ -73,8 +89,13 @@ def LoadDataFrame(df, x_col=0, y_col=1, name=None):
         x_col = [df.columns[item] for item in x_col]
     if all(isinstance(item, int) for item in y_col):
         y_col = [df.columns[item] for item in y_col]
+    if y_err_col is not None and all(isinstance(item, int) for item in y_err_col):
+        y_err_col = [df.columns[item] for item in y_err_col]
 
-    df = df[x_col + y_col]
+    cols = x_col + y_col
+    if y_err_col is not None:
+        cols += y_err_col
+    df = df[cols]
     if len(df.index) == 0:
         raise ValueError("dataframe cannot be empty")
 
@@ -84,11 +105,19 @@ def LoadDataFrame(df, x_col=0, y_col=1, name=None):
 
     dataset = DataSet()
     for i in range(len(y_col)):
-        channel = df[x_col + [y_col[i]]].dropna()
+        cols = x_col + [y_col[i]]
+        if y_err_col is not None:
+            cols += [y_err_col[i]]
+        channel = df[cols].dropna()
+
+        y_err = None
+        if y_err_col is not None:
+            y_err = channel[y_err_col[i]].values
 
         dataset.append(Data(
             channel[x_col].values,
             channel[y_col[i]].values,
+            Y_err=y_err,
             name=name[i],
             x_labels=x_labels,
             y_label=str(y_col[i]),
@@ -151,6 +180,27 @@ class DataSet:
         for arg in args:
             self.append(arg)
 
+    def _format_X(self, X):
+        if isinstance(X, dict):
+            x_dict = X
+            X = self.get_prediction()
+            for name, channel_x in x_dict.items():
+                X[self.get_index(name)] = channel_x
+        elif isinstance(X, np.ndarray):
+            X = [X] * self.get_output_dims()
+        elif isinstance(X, pd.Series):
+            X = [X.to_numpy()] * self.get_output_dims()
+        elif not isinstance(X, list):
+            raise ValueError("X must be a list, dict, numpy.ndarray, or pandas.Series")
+        elif not any(isinstance(x, (list,np.ndarray)) for x in X):
+            X = [X] * self.get_output_dims()
+        if len(X) != self.get_output_dims():
+            raise ValueError("X must be of shape (data_points,), (data_points,input_dims), or [(data_points,)] * input_dims for each channel")
+
+        for j, channel in enumerate(self.channels):
+            X[j], _ = channel._format_X(X[j])
+        return X
+
     def __iter__(self):
         return self.channels.__iter__()
 
@@ -217,19 +267,6 @@ class DataSet:
         """
         return copy.deepcopy(self)
 
-    def rescale_x(self, upper=1000.0):
-        """
-        Rescale the X axis so that it is in the interval [0.0,upper] for each channel. This helps training most kernels.
-
-        Args:
-            upper (float): Upper end of the interval.
-
-        Examples:
-            >>> dataset.rescale_x()
-        """
-        for channel in self.channels:
-            channel.rescale_x(upper)
-
     def transform(self, transformer):
         """
         Transform each channel by using one of the provided transformers, such as `TransformDetrend`, `TransformLinear`, `TransformLog`, `TransformNormalize`, `TransformStandard`, etc.
@@ -247,13 +284,14 @@ class DataSet:
         for channel in self.channels:
             channel.transform(transformer)
 
-    def filter(self, start, end):
+    def filter(self, start, end, dim=None):
         """
         Filter the data range to be between `start` and `end` in the X axis.
 
         Args:
-            start (float, str): Start of interval.
-            end (float, str): End of interval.
+            start (float, str, list): Start of interval.
+            end (float, str, list): End of interval.
+            dim (int): Input dimension to apply to, if not specified applies to all input dimensions.
 
         Examples:
             >>> dataset.filter(3, 8)
@@ -261,7 +299,7 @@ class DataSet:
             >>> dataset.filter('2016-01-15', '2016-06-15')
         """
         for channel in self.channels:
-            channel.filter(start, end)
+            channel.filter(start, end, dim=dim)
 
     def aggregate(self, duration, f=np.mean):
         """
@@ -281,6 +319,19 @@ class DataSet:
         for channel in self.channels:
             channel.aggregate(duration, f)
 
+    def has_test_data(self):
+        """
+        Returns True if observations have been removed using the `remove_*` methods.
+
+        Returns:
+            list: Boolean per channel.
+
+        Examples:
+            >>> data.has_test_data()
+            True
+        """
+        return [channel.has_test_data() for channel in self.channels]
+
     def get_input_dims(self):
         """
         Return the input dimensions per channel.
@@ -372,8 +423,8 @@ class DataSet:
             transformed (boolean): Return transformed data.
 
         Returns:
-            list: X data of shape (n,input_dims) per channel.
-            list: Y data of shape (n,) per channel.
+            list: X data of shape (data_points,input_dims) per channel.
+            list: Y data of shape (data_points,) per channel.
 
         Examples:
             >>> x, y = dataset.get_data()
@@ -388,8 +439,8 @@ class DataSet:
             transformed (boolean): Return transformed data.
 
         Returns:
-            list: X data of shape (n,input_dims) per channel.
-            list: Y data of shape (n,) per channel.
+            list: X data of shape (data_points,input_dims) per channel.
+            list: Y data of shape (data_points,) per channel.
 
         Examples:
             >>> x, y = dataset.get_train_data()
@@ -404,82 +455,52 @@ class DataSet:
             transformed (boolean): Return transformed data.
 
         Returns:
-            list: X data of shape (n,input_dims) per channel.
-            list: Y data of shape (n,) per channel.
+            list: X data of shape (data_points,input_dims) per channel.
+            list: Y data of shape (data_points,) per channel.
 
         Examples:
             >>> x, y = dataset.get_test_data()
         """
         return [channel.get_test_data(transformed=transformed)[0] for channel in self.channels], [channel.get_test_data(transformed=transformed)[1] for channel in self.channels]
-    
-    def get_prediction_x(self):
+
+    def get_prediction_data(self):
         """
         Returns the prediction X range for all channels.
 
         Returns:
-            list: X prediction of shape (n,input_dims) per channel.
+            list: X prediction of shape (data_points,input_dims) per channel.
 
         Examples:
-            >>> x = dataset.get_prediction_x()
+            >>> x = dataset.get_prediction_data()
         """
         x = []
         for channel in self.channels:
-            x.append(channel.get_prediction_x())
+            x.append(channel.get_prediction_data())
         return x
-    
-    def get_prediction(self, name, sigma=2.0, transformed=False):
-        """
-        Returns the prediction of a given name with a confidence interval of `sigma` times the standard deviation.
 
-        Args:
-            name (str): Name of the model of the prediction.
-            sigma (float): The confidence interval's number of standard deviations.
-            transformed (boolean): Return transformed data as used for training.
-
-        Returns:
-            list: X prediction of shape (n,input_dims) per channel.
-            list: Y mean prediction of shape (n,) per channel.
-            list: Y lower prediction of uncertainty interval of shape (n,) per channel.
-            list: Y upper prediction of uncertainty interval of shape (n,) per channel.
-
-        Examples:
-            >>> x, y_mean, y_var_lower, y_var_upper = dataset.get_prediction('MOSM', sigma=1)
-        """
-        x = []
-        mu = []
-        lower = []
-        upper = []
-        for channel in self.channels:
-            cx, cmu, clower, cupper = channel.get_prediction(name, sigma, transformed=transformed)
-            x.append(cx)
-            mu.append(cmu)
-            lower.append(clower)
-            upper.append(cupper)
-        return x, mu, lower, upper
-
-    def set_prediction_x(self, x):
+    def set_prediction_data(self, X):
         """
         Set the prediction range directly for saved predictions per channel. This will clear old predictions.
 
         Args:
-            x (list, dict): Array of shape (n,) or (n,input_dims) per channel with prediction X values. If a dictionary is passed, the index is the channel index or name.
+            X (list, dict): Array of shape (data_points,), (data_points,input_dims), or [(data_points,)] * input_dims per channel with prediction X values. If a dictionary is passed, the index is the channel index or name.
 
         Examples:
-            >>> dataset.set_prediction_x([[5.0, 5.5, 6.0, 6.5, 7.0], [0.1, 0.2, 0.3]])
-            >>> dataset.set_prediction_x({'A': [5.0, 5.5, 6.0, 6.5, 7.0], 'B': [0.1, 0.2, 0.3]})
+            >>> dataset.set_prediction_data([[5.0, 5.5, 6.0, 6.5, 7.0], [0.1, 0.2, 0.3]])
+            >>> dataset.set_prediction_data({'A': [5.0, 5.5, 6.0, 6.5, 7.0], 'B': [0.1, 0.2, 0.3]})
         """
-        if isinstance(x, list):
-            if len(x) != len(self.channels):
+        if isinstance(X, list):
+            if len(X) != len(self.channels):
                 raise ValueError("prediction x expected to be a list of shape (output_dims,n)")
 
             for i, channel in enumerate(self.channels):
-                channel.set_prediction_x(x[i])
-        elif isinstance(x, dict):
-            for name in x:
-                self.get(name).set_prediction_x(x[name])
+                channel.set_prediction_data(X[i])
+        elif isinstance(X, dict):
+            for name in X:
+                self.get(name).set_prediction_data(X[name])
         else:
             for i, channel in enumerate(self.channels):
-                channel.set_prediction_x(x)
+                channel.set_prediction_data(X)
 
     def set_prediction_range(self, start, end, n=None, step=None):
         """
@@ -521,13 +542,6 @@ class DataSet:
 
         for i, channel in enumerate(self.channels):
             channel.set_prediction_range(start[i], end[i], n[i], step[i])
-
-    def clear_predictions(self):
-        """
-        Clear all saved predictions for all channels.
-        """
-        for i, channel in enumerate(self.channels):
-            channel.clear_predictions()
     
     def get_nyquist_estimation(self):
         """
@@ -541,7 +555,7 @@ class DataSet:
         """
         return [channel.get_nyquist_estimation() for channel in self.channels]
     
-    def get_lombscargle_estimation(self, Q=1, n=10000):
+    def get_ls_estimation(self, Q=1, n=10000):
         """
         Peak estimation of the spectrum using Lomb-Scargle per channel.
 
@@ -561,19 +575,20 @@ class DataSet:
         means = []
         variances = []
         for channel in self.channels:
-            channel_amplitudes, channel_means, channel_variances = channel.get_lombscargle_estimation(Q, n)
+            channel_amplitudes, channel_means, channel_variances = channel.get_ls_estimation(Q, n)
             amplitudes.append(channel_amplitudes)
             means.append(channel_means)
             variances.append(channel_variances)
         return amplitudes, means, variances
     
-    def get_bnse_estimation(self, Q=1, n=1000):
+    def get_bnse_estimation(self, Q=1, n=1000, iters=200):
         """
         Peak estimation of the spectrum using BNSE (Bayesian Non-parametric Spectral Estimation) per channel.
 
         Args:
             Q (int): Number of peaks to find.
             n (int): Number of points of the grid to evaluate frequencies.
+            iters (str): Maximum iterations.
 
         Returns:
             list: Amplitude array of shape (Q,input_dims) per channel.
@@ -587,23 +602,22 @@ class DataSet:
         means = []
         variances = []
         for channel in self.channels:
-            channel_amplitudes, channel_means, channel_variances = channel.get_bnse_estimation(Q, n)
+            channel_amplitudes, channel_means, channel_variances = channel.get_bnse_estimation(Q, n, iters=iters)
             amplitudes.append(channel_amplitudes)
             means.append(channel_means)
             variances.append(channel_variances)
         return amplitudes, means, variances
     
-    def get_sm_estimation(self, Q=1, method='BNSE', optimizer='Adam', iters=100, params={}, plot=False):
+    def get_sm_estimation(self, Q=1, method='BNSE', optimizer='Adam', iters=200, params={}):
         """
         Peak estimation of the spectrum using the spectral mixture kernel per channel.
 
         Args:
             Q (int): Number of peaks to find.
-            method (str): Method of estimating SM kernels.
-            optimizer (str): Optimization method for SM kernels.
-            iters (str): Maximum iteration for SM kernels.
+            method (str): Method of estimation.
+            optimizer (str): Optimization method.
+            iters (str): Maximum iterations.
             params (object): Additional parameters for PyTorch optimizer.
-            plot (bool): Show the PSD of the kernel after fitting.
 
         Returns:
             list: Amplitude array of shape (Q,input_dims) per channel.
@@ -617,7 +631,7 @@ class DataSet:
         means = []
         variances = []
         for channel in self.channels:
-            channel_amplitudes, channel_means, channel_variances = channel.get_sm_estimation(Q, method, optimizer, iters, params, plot)
+            channel_amplitudes, channel_means, channel_variances = channel.get_sm_estimation(Q, method, optimizer, iters, params)
             amplitudes.append(channel_amplitudes)
             means.append(channel_means)
             variances.append(channel_variances)
@@ -642,7 +656,7 @@ class DataSet:
             >>> fig, axes = dataset.plot(title='Title')
         """
         if figsize is None:
-            figsize = (12, 3.0 * len(self.channels))
+            figsize = (12,4*len(self.channels))
 
         h = figsize[1]
         fig, axes = plt.subplots(self.get_output_dims(), 1, figsize=figsize, squeeze=False, constrained_layout=True)
@@ -650,21 +664,22 @@ class DataSet:
         legends = {}
         for channel in range(self.get_output_dims()):
             self.channels[channel].plot(pred=pred, ax=axes[channel,0], transformed=transformed)
-            legend = axes[channel,0].get_legend()
-            for text, handle in zip(legend.texts, legend.legendHandles):
-                if text.get_text() == "Training Points":
-                    handle = plt.Line2D([0], [0], ls='-', color='k', marker='.', ms=10, label='Training Points')
+            l = axes[channel,0].get_legend()
+            for text, handle in zip(l.texts, l.legendHandles):
+                if text.get_text() == "Observations":
+                    handle = plt.Line2D([0], [0], ls='', color='r', marker='.', ms=10, label='Observations')
                 legends[text.get_text()] = handle
-            legend.remove()
+            l.remove()
 
         legend_rows = (len(legends)-1)/5 + 1
         if title is not None:
             fig.suptitle(title, y=(h+0.2+0.4*legend_rows)/h, fontsize=18)
 
-        fig.legend(handles=legends.values(), loc="upper center", bbox_to_anchor=(0.5,(h-0.3+0.5*legend_rows)/h), ncol=5)
+        if legend:
+            fig.legend(handles=legends.values(), ncol=5)
         return fig, axes
 
-    def plot_spectrum(self, title=None, method='ls', per=None, maxfreq=None, figsize=None, transformed=False):
+    def plot_spectrum(self, title=None, method='ls', per=None, maxfreq=None, figsize=None, log=False, transformed=True, n=1001):
         """
         Plot the spectrum for each channel.
 
@@ -674,7 +689,9 @@ class DataSet:
             per (list, str): Set the scale of the X axis depending on the formatter used, eg. per=5, per='day', or per='3D'.
             maxfreq (list, float): Maximum frequency to plot, otherwise the Nyquist frequency is used.
             figsize (tuple): Set the figure size.
+            log (boolean): Show X and Y axis in log-scale.
             transformed (boolean): Display transformed Y data as used for training.
+            n (int): Number of points used for periodogram.
 
         Returns:
             matplotlib.figure.Figure: The figure.
@@ -691,12 +708,12 @@ class DataSet:
             maxfreq = [maxfreq] * len(self.channels)
 
         if figsize is None:
-            figsize = (12, 3.0 * len(self.channels))
+            figsize = (12,4*len(self.channels))
 
         fig, axes = plt.subplots(self.get_output_dims(), 1, figsize=figsize, squeeze=False, constrained_layout=True)
         if title != None:
             fig.suptitle(title, fontsize=18)
 
         for channel in range(self.get_output_dims()):
-            ax = self.channels[channel].plot_spectrum(method=method[channel], ax=axes[channel,0], per=per[channel], maxfreq=maxfreq[channel], transformed=transformed)
+            self.channels[channel].plot_spectrum(method=method[channel], ax=axes[channel,0], per=per[channel], maxfreq=maxfreq[channel], log=log, transformed=transformed, n=n)
         return fig, axes
